@@ -2,6 +2,7 @@ import argparse
 import logging
 import asyncio
 import yaml
+import signal
 import sys
 
 import ssl
@@ -26,6 +27,7 @@ from dnstap_receiver.outputs import output_syslog
 from dnstap_receiver.outputs import output_tcp
 from dnstap_receiver.outputs import output_metrics
 from dnstap_receiver.outputs import output_dnstap
+from dnstap_receiver.outputs import output_kafka
 
 from dnstap_receiver import api_server
 from dnstap_receiver import statistics
@@ -47,6 +49,7 @@ parser.add_argument("-c", help="external config file")
 
 # get event loop
 loop = asyncio.get_event_loop()
+shutdown_task = None
 
 def merge_cfg(u, o):
     """merge config"""
@@ -119,7 +122,7 @@ def setup_logger(cfg):
     
     clogger.addHandler(lh)
     
-def setup_outputs(cfg, stats):
+def setup_outputs(cfg, stats, start_shutdown):
     """setup outputs"""
     conf = cfg["output"]
 
@@ -128,62 +131,68 @@ def setup_outputs(cfg, stats):
         if not output_syslog.checking_conf(cfg=conf["syslog"]): return
         queue_syslog = asyncio.Queue()
         queues_list.append(queue_syslog)
-        loop.create_task(output_syslog.handle(conf["syslog"], queue_syslog, stats))    
+        loop.create_task(output_syslog.handle(conf["syslog"], queue_syslog, stats, start_shutdown))
 
     if conf["tcp-socket"]["enable"]:
         if not output_tcp.checking_conf(cfg=conf["tcp-socket"]): return
         queue_tcpsocket = asyncio.Queue()
         queues_list.append(queue_tcpsocket)
-        loop.create_task(output_tcp.handle(conf["tcp-socket"], queue_tcpsocket, stats))
+        loop.create_task(output_tcp.handle(conf["tcp-socket"], queue_tcpsocket, stats, start_shutdown))
                                                
     if conf["file"]["enable"]:
         if not output_file.checking_conf(cfg=conf["file"]): return
         queue_file = asyncio.Queue()
         queues_list.append(queue_file)
-        loop.create_task(output_file.handle(conf["file"], queue_file, stats))
+        loop.create_task(output_file.handle(conf["file"], queue_file, stats, start_shutdown))
                                               
     if conf["stdout"]["enable"]:
         if not output_stdout.checking_conf(cfg=conf["stdout"]): return
         queue_stdout = asyncio.Queue()
         queues_list.append(queue_stdout)
-        loop.create_task(output_stdout.handle(conf["stdout"], queue_stdout, stats))
+        loop.create_task(output_stdout.handle(conf["stdout"], queue_stdout, stats, start_shutdown))
 
     if conf["metrics"]["enable"]:
         if not output_metrics.checking_conf(cfg=conf["metrics"]): return
         queue_metrics = asyncio.Queue()
         queues_list.append(queue_metrics)
-        loop.create_task(output_metrics.handle(conf["metrics"], queue_metrics, stats))
+        loop.create_task(output_metrics.handle(conf["metrics"], queue_metrics, stats, start_shutdown))
 
     if conf["dnstap"]["enable"]:
         if not output_dnstap.checking_conf(cfg=conf["dnstap"]): return
         queue_dnstap = asyncio.Queue()
         queues_list.append(queue_dnstap)
-        loop.create_task(output_dnstap.handle(conf["dnstap"], queue_dnstap, stats))
-        
+        loop.create_task(output_dnstap.handle(conf["dnstap"], queue_dnstap, stats, start_shutdown))
+
+    if conf["kafka"]["enable"]:
+        if not output_kafka.checking_conf(cfg=conf["kafka"]): return
+        queue_kafka = asyncio.Queue()
+        queues_list.append(queue_kafka)
+        loop.create_task(output_kafka.handle(conf["kafka"], queue_kafka, stats, start_shutdown))
+
     return queues_list
     
-def setup_inputs(cfg, queues_outputs, stats, geoip_reader, running):
+def setup_inputs(cfg, queues_outputs, stats, geoip_reader, start_shutdown):
     """setup inputs"""
     cache = cachetools.TTLCache(maxsize=1000000, ttl=60)
     cfg_input = cfg["input"]
     
     # asynchronous unix 
     if cfg_input["unix-socket"]["enable"]:
-        loop.create_task(input_socket.start_unixsocket(cfg, queues_outputs, stats, geoip_reader, cache)) 
+        loop.create_task(input_socket.start_unixsocket(cfg, queues_outputs, stats, geoip_reader, cache, start_shutdown))
         
     # sniffer
     elif cfg_input["sniffer"]["enable"]:
         queue_sniffer = asyncio.Queue()
-        loop.create_task(input_sniffer.watch_buffer(cfg_input["sniffer"], queue_sniffer, queues_outputs, stats, cache))
-        loop.run_in_executor(None, input_sniffer.start_input, cfg_input["sniffer"], queue_sniffer, running)
+        loop.create_task(input_sniffer.watch_buffer(cfg_input["sniffer"], queue_sniffer, queues_outputs, stats, cache, start_shutdown))
+        loop.run_in_executor(None, input_sniffer.start_input, cfg_input["sniffer"], queue_sniffer, start_shutdown)
     
     # tcp client input
     elif cfg_input["tcp-client"]["enable"]:
-        loop.create_task(input_socket.start_tcpclient(cfg, queues_outputs, stats, geoip_reader, cache))
+        loop.create_task(input_socket.start_tcpclient(cfg, queues_outputs, stats, geoip_reader, cache, start_shutdown))
         
     # default one tcp socket
     else:
-        loop.create_task(input_socket.start_tcpsocket(cfg, queues_outputs, stats, geoip_reader, cache))
+        loop.create_task(input_socket.start_tcpsocket(cfg, queues_outputs, stats, geoip_reader, cache, start_shutdown))
 
 def setup_webserver(cfg, stats):
     """setup web api"""
@@ -205,7 +214,34 @@ def setup_geoip(cfg):
         clogger.error("geoip setup: %s" % e)
         
     return reader
-    
+
+def start_shutdown_task(signal, loop, start_shutdown):
+    global shutdown_task
+    if not shutdown_task:
+        shutdown_task = asyncio.create_task(
+            shutdown(signal, loop, start_shutdown),
+            name='shutdown',
+        )
+
+async def shutdown(signal, loop, start_shutdown):
+    """perform graceful shutdown"""
+
+    clogger.info("starting shutting down process")
+    start_shutdown.set()
+
+    current_task = asyncio.current_task()
+    tasks = [
+        task for task in asyncio.all_tasks()
+        if task is not current_task
+    ]
+
+    clogger.info("waiting for all tasks to exit")
+    await asyncio.gather(*tasks, return_exceptions=True)
+
+    clogger.debug("all tasks have exited, stopping loop")
+    loop.stop()
+
+
 def start_receiver():
     """start dnstap receiver"""
     # Handle command-line arguments.
@@ -220,20 +256,24 @@ def start_receiver():
     # setup geoip if enabled 
     geoip_reader = setup_geoip(cfg=cfg["geoip"])
     
+    # prepare shutdown handling
+    start_shutdown = asyncio.Event()
+    for sig in (signal.SIGHUP, signal.SIGTERM, signal.SIGINT):
+        loop.add_signal_handler(sig, lambda: start_shutdown_task(sig, loop, start_shutdown))
+
     # add debug message if external config is used
     if args.c: clogger.debug("External config file loaded")
     
     # start receiver
     clogger.debug("Start receiver...")
     stats = statistics.Statistics(cfg=cfg["statistics"])
-    loop.create_task(statistics.watcher(stats))
+    loop.create_task(statistics.watcher(stats, start_shutdown))
     
     # prepare outputs
-    queues_outputs = setup_outputs(cfg, stats)
-    
+    queues_outputs = setup_outputs(cfg, stats, start_shutdown)
+
     # prepare inputs
-    running = ["running"]
-    setup_inputs(cfg, queues_outputs, stats, geoip_reader, running)
+    setup_inputs(cfg, queues_outputs, stats, geoip_reader, start_shutdown)
 
     # start the http api
     setup_webserver(cfg, stats)
@@ -241,9 +281,10 @@ def start_receiver():
     # run event loop 
     try:
        loop.run_forever()
-    except KeyboardInterrupt:
-        clogger.debug("exiting, please wait..")
-        running.clear()
+    finally:
+        clogger.info("exiting, please wait..")
+        loop.close()
+        clogger.debug("shut down eventloop")
 
     # close geoip
-    if geoip_reader is not None: geoip_reader.close()
\ No newline at end of file
+    if geoip_reader is not None: geoip_reader.close()
